@@ -16,14 +16,20 @@
   limitations under the License.
 */
 
-#include "tscore/ink_platform.h"
 #include <netinet/in.h>
+#include <arpa/inet.h>
 #include "ts_lua_util.h"
 
 static int ts_lua_get_process_id(lua_State *L);
 static int ts_lua_get_now_time(lua_State *L);
 static int ts_lua_debug(lua_State *L);
 static int ts_lua_error(lua_State *L);
+static int ts_lua_emergency(lua_State *L);
+static int ts_lua_fatal(lua_State *L);
+static int ts_lua_status(lua_State *L);
+static int ts_lua_note(lua_State *L);
+static int ts_lua_warning(lua_State *L);
+static int ts_lua_alert(lua_State *L);
 static int ts_lua_sleep(lua_State *L);
 static int ts_lua_host_lookup(lua_State *L);
 static int ts_lua_schedule(lua_State *L);
@@ -64,6 +70,30 @@ ts_lua_inject_misc_api(lua_State *L)
   lua_pushcfunction(L, ts_lua_error);
   lua_setfield(L, -2, "error");
 
+  /* ts.emergency(...) */
+  lua_pushcfunction(L, ts_lua_emergency);
+  lua_setfield(L, -2, "emergency");
+
+  /* ts.fatal(...) */
+  lua_pushcfunction(L, ts_lua_fatal);
+  lua_setfield(L, -2, "fatal");
+
+  /* ts.status(...) */
+  lua_pushcfunction(L, ts_lua_status);
+  lua_setfield(L, -2, "status");
+
+  /* ts.note(...) */
+  lua_pushcfunction(L, ts_lua_note);
+  lua_setfield(L, -2, "note");
+
+  /* ts.warning(...) */
+  lua_pushcfunction(L, ts_lua_warning);
+  lua_setfield(L, -2, "warning");
+
+  /* ts.alert(...) */
+  lua_pushcfunction(L, ts_lua_alert);
+  lua_setfield(L, -2, "alert");
+
   /* ts.sleep(...) */
   lua_pushcfunction(L, ts_lua_sleep);
   lua_setfield(L, -2, "sleep");
@@ -163,6 +193,72 @@ ts_lua_error(lua_State *L)
   return 0;
 }
 
+static int
+ts_lua_emergency(lua_State *L)
+{
+  const char *msg;
+  size_t len = 0;
+
+  msg = luaL_checklstring(L, 1, &len);
+  TSEmergency("%.*s", (int)len, msg);
+  return 0;
+}
+
+static int
+ts_lua_fatal(lua_State *L)
+{
+  const char *msg;
+  size_t len = 0;
+
+  msg = luaL_checklstring(L, 1, &len);
+  TSFatal("%.*s", (int)len, msg);
+  return 0;
+}
+
+static int
+ts_lua_status(lua_State *L)
+{
+  const char *msg;
+  size_t len = 0;
+
+  msg = luaL_checklstring(L, 1, &len);
+  TSStatus("%.*s", (int)len, msg);
+  return 0;
+}
+
+static int
+ts_lua_note(lua_State *L)
+{
+  const char *msg;
+  size_t len = 0;
+
+  msg = luaL_checklstring(L, 1, &len);
+  TSNote("%.*s", (int)len, msg);
+  return 0;
+}
+
+static int
+ts_lua_warning(lua_State *L)
+{
+  const char *msg;
+  size_t len = 0;
+
+  msg = luaL_checklstring(L, 1, &len);
+  TSWarning("%.*s", (int)len, msg);
+  return 0;
+}
+
+static int
+ts_lua_alert(lua_State *L)
+{
+  const char *msg;
+  size_t len = 0;
+
+  msg = luaL_checklstring(L, 1, &len);
+  TSAlert("%.*s", (int)len, msg);
+  return 0;
+}
+
 static int
 ts_lua_schedule(lua_State *L)
 {
@@ -180,8 +276,6 @@ ts_lua_schedule(lua_State *L)
 
   ci = ts_lua_get_cont_info(L);
   if (ci == NULL) {
-    TSError("[ts_lua][%s] no cont info found", __FUNCTION__);
-    TSReleaseAssert(!"Unexpected fetch of cont info");
     return 0;
   }
 
@@ -199,7 +293,7 @@ ts_lua_schedule(lua_State *L)
   n = lua_gettop(L);
 
   if (n < 3) {
-    TSError("[ts_lua] ts.schedule need at least three parameters");
+    TSError("[ts_lua][%s] ts.schedule need at least three parameters", __FUNCTION__);
     return 0;
   }
 
@@ -213,7 +307,7 @@ ts_lua_schedule(lua_State *L)
   nci->contp = contp;
   nci->mutex = ci->mutex;
 
-  TSContSchedule(contp, sec * 1000, entry);
+  TSContScheduleOnPool(contp, sec * 1000, entry);
 
   return 0;
 }
@@ -258,7 +352,7 @@ ts_lua_schedule_handler(TSCont contp, TSEvent ev, void *edata)
   }
 
   if (ret != 0) {
-    TSError("[ts_lua] lua_resume failed: %s", lua_tostring(L, -1));
+    TSError("[ts_lua][%s] lua_resume failed: %s", __FUNCTION__, lua_tostring(L, -1));
   }
 
   lua_pop(L, lua_gettop(L));
@@ -280,8 +374,6 @@ ts_lua_sleep(lua_State *L)
 
   ci = ts_lua_get_cont_info(L);
   if (ci == NULL) {
-    TSError("[ts_lua][%s] no cont info found", __FUNCTION__);
-    TSReleaseAssert(!"Unexpected fetch of cont info");
     return 0;
   }
 
@@ -291,7 +383,7 @@ ts_lua_sleep(lua_State *L)
   }
 
   contp  = TSContCreate(ts_lua_sleep_handler, ci->mutex);
-  action = TSContSchedule(contp, sec * 1000, TS_THREAD_POOL_DEFAULT);
+  action = TSContScheduleOnPool(contp, sec * 1000, TS_THREAD_POOL_NET);
 
   ai = ts_lua_async_create_item(contp, ts_lua_sleep_cleanup, (void *)action, ci);
   TSContDataSet(contp, ai);
@@ -342,13 +434,11 @@ ts_lua_host_lookup(lua_State *L)
 
   ci = ts_lua_get_cont_info(L);
   if (ci == NULL) {
-    TSError("[ts_lua][%s] no cont info found", __FUNCTION__);
-    TSReleaseAssert(!"Unexpected fetch of cont info");
     return 0;
   }
 
   if (lua_gettop(L) != 1) {
-    TSError("[ts_lua] ts.host_lookup need at least one parameter");
+    TSError("[ts_lua][%s] ts.host_lookup need at least one parameter", __FUNCTION__);
     return 0;
   }
 
@@ -382,7 +472,7 @@ ts_lua_host_lookup_handler(TSCont contp, TSEvent event, void *edata)
   ts_lua_host_lookup_cleanup(ai);
 
   if (event != TS_EVENT_HOST_LOOKUP) {
-    TSError("[ts_lua] ts.host_lookup receives unknown event");
+    TSError("[ts_lua][%s] ts.host_lookup receives unknown event", __FUNCTION__);
     lua_pushnil(L);
   } else if (!edata) {
     lua_pushnil(L);
